from typing import Callable, Optional, Sequence, Tuple, Type

import cached_conv as cc
import numpy as np
import torch
import torch.nn as nn
import torchaudio

from .blocks import normalization


def spectrogram(n_fft: int):
    return torchaudio.transforms.Spectrogram(
        n_fft,
        hop_length=n_fft // 4,
        power=None,
        normalized=True,
        center=False,
        pad_mode=None,
    )


def rectified_2d_conv_block(
    capacity,
    kernel_sizes,
    strides: Optional[Tuple[int, int]] = None,
    dilations: Optional[Tuple[int, int]] = None,
    in_size: Optional[int] = None,
    out_size: Optional[int] = None,
    activation: bool = True,
):
    if dilations is None:
        paddings = kernel_sizes[0] // 2, kernel_sizes[1] // 2
    else:
        fks = (kernel_sizes[0] - 1) * dilations[0], (kernel_sizes[1] -
                                                     1) * dilations[1]
        paddings = fks[0] // 2, fks[1] // 2

    conv = normalization(
        nn.Conv2d(
            in_size or capacity,
            out_size or capacity,
            kernel_size=kernel_sizes,
            stride=strides or (1, 1),
            dilation=dilations or (1, 1),
            padding=paddings,
        ))

    if not activation: return conv

    return nn.Sequential(conv, nn.LeakyReLU(.2))


class EncodecConvNet(nn.Module):

    def __init__(self, capacity: int) -> None:
        super().__init__()
        self.net = nn.Sequential(
            rectified_2d_conv_block(capacity, (9, 3), in_size=2),
            rectified_2d_conv_block(capacity, (9, 3), (2, 1), (1, 1)),
            rectified_2d_conv_block(capacity, (9, 3), (2, 1), (1, 2)),
            rectified_2d_conv_block(capacity, (9, 3), (2, 1), (1, 4)),
            rectified_2d_conv_block(capacity, (3, 3)),
            rectified_2d_conv_block(capacity, (3, 3),
                                    out_size=1,
                                    activation=False),
        )

    def forward(self, x):
        features = []
        for layer in self.net:
            x = layer(x)
            features.append(x)
        return features


class ConvNet(nn.Module):

    def __init__(self, in_size, out_size, capacity, n_layers, kernel_size,
                 stride, conv) -> None:
        super().__init__()
        channels = [in_size]
        channels += list(capacity * 2**np.arange(n_layers))

        if isinstance(stride, int):
            stride = n_layers * [stride]

        net = []
        for i in range(n_layers):
            if not isinstance(kernel_size, int):
                pad = (cc.get_padding(kernel_size[0],
                                      stride[i],
                                      mode="centered")[0], 0)
                s = (stride[i], 1)
            else:
                pad = cc.get_padding(kernel_size, stride[i],
                                     mode="centered")[0]
                s = stride[i]
            net.append(
                normalization(
                    conv(
                        channels[i],
                        channels[i + 1],
                        kernel_size,
                        stride=s,
                        padding=pad,
                    )))
            net.append(nn.LeakyReLU(.2))
        net.append(conv(channels[-1], out_size, 1))

        self.net = nn.Sequential(*net)

    def forward(self, x):
        features = []
        for layer in self.net:
            x = layer(x)
            if isinstance(layer, nn.modules.conv._ConvNd):
                features.append(x)
        return features


class MultiScaleDiscriminator(nn.Module):

    def __init__(self, n_discriminators, convnet) -> None:
        super().__init__()
        layers = []
        for i in range(n_discriminators):
            layers.append(convnet())
        self.layers = nn.ModuleList(layers)

    def forward(self, x):
        features = []
        for layer in self.layers:
            features.append(layer(x))
            x = nn.functional.avg_pool1d(x, 2)
        return features


class MultiScaleSpectralDiscriminator(nn.Module):

    def __init__(self, scales: Sequence[int],
                 convnet: Callable[[], nn.Module]) -> None:
        super().__init__()
        self.specs = nn.ModuleList([spectrogram(n) for n in scales])
        self.nets = nn.ModuleList([convnet() for _ in scales])

    def forward(self, x):
        features = []
        for spec, net in zip(self.specs, self.nets):
            spec_x = spec(x)
            spec_x = torch.cat([spec_x.real, spec_x.imag], 1)
            features.append(net(spec_x))
        return features


class MultiScaleSpectralDiscriminator1d(nn.Module):

    def __init__(self, scales: Sequence[int],
                 convnet: Callable[[int], nn.Module]) -> None:
        super().__init__()
        self.specs = nn.ModuleList([spectrogram(n) for n in scales])
        self.nets = nn.ModuleList([convnet(n + 2) for n in scales])

    def forward(self, x):
        features = []
        for spec, net in zip(self.specs, self.nets):
            spec_x = spec(x).squeeze(1)
            spec_x = torch.cat([spec_x.real, spec_x.imag], 1)
            features.append(net(spec_x))
        return features


class MultiPeriodDiscriminator(nn.Module):

    def __init__(self, periods, convnet) -> None:
        super().__init__()
        layers = []
        self.periods = periods

        for _ in periods:
            layers.append(convnet())

        self.layers = nn.ModuleList(layers)

    def forward(self, x):
        features = []
        for layer, n in zip(self.layers, self.periods):
            features.append(layer(self.fold(x, n)))
        return features

    def fold(self, x, n):
        pad = (n - (x.shape[-1] % n)) % n
        x = nn.functional.pad(x, (0, pad))
        return x.reshape(*x.shape[:2], -1, n)


class CombineDiscriminators(nn.Module):

    def __init__(self, discriminators: Sequence[Type[nn.Module]]) -> None:
        super().__init__()
        self.discriminators = nn.ModuleList(disc_cls()
                                            for disc_cls in discriminators)

    def forward(self, x):
        features = []
        for disc in self.discriminators:
            features.extend(disc(x))
        return features



import gin
import pytest
import torch
import torch.nn as nn

import rave
from scripts import export

gin.enter_interactive_mode()

configs = [
    ["v1.gin"],
    ["v2.gin"],
    ["v2.gin", "wasserstein.gin"],
    ["v2.gin", "spherical.gin"],
    ["discrete.gin"],
    ["discrete.gin", "spectral_discriminator.gin"],
]

configs += [c + ["causal.gin"] for c in configs]


@pytest.mark.parametrize("config", configs, ids=lambda elm: " ".join(elm))
def test_config(config):
    gin.clear_config()
    gin.parse_config_files_and_bindings(config, [])

    model = rave.RAVE()

    x = torch.randn(1, 1, 2**15)
    z = model.encode(x)
    y = model.decode(z)
    score = model.discriminator(y)

    assert x.shape == y.shape

    if isinstance(model.encoder, rave.blocks.VariationalEncoder):
        script_class = export.VariationalScriptedRAVE
    elif isinstance(model.encoder, rave.blocks.DiscreteEncoder):
        script_class = export.DiscreteScriptedRAVE
    elif isinstance(model.encoder, rave.blocks.WasserteinEncoder):
        script_class = export.WasserteinScriptedRAVE
    elif isinstance(model.encoder, rave.blocks.SphericalEncoder):
        script_class = export.SphericalScriptedRAVE
    else:
        raise ValueError(f"Encoder type {type(model.encoder)} "
                         "not supported for export.")

    x = torch.zeros(1, 1, 2**14)
    model(x)

    for m in model.modules():
        if hasattr(m, "weight_g"):
            nn.utils.remove_weight_norm(m)

    scripted_rave = script_class(pretrained=model, stereo=False)



import math

import cached_conv as cc
import gin
import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from scipy.optimize import fmin
from scipy.signal import firwin, kaiser, kaiser_beta, kaiserord


def reverse_half(x):
    mask = torch.ones_like(x)
    mask[..., 1::2, ::2] = -1

    return x * mask


def center_pad_next_pow_2(x):
    next_2 = 2**math.ceil(math.log2(x.shape[-1]))
    pad = next_2 - x.shape[-1]
    return nn.functional.pad(x, (pad // 2, pad // 2 + int(pad % 2)))


def make_odd(x):
    if not x.shape[-1] % 2:
        x = nn.functional.pad(x, (0, 1))
    return x


def get_qmf_bank(h, n_band):
    """
    Modulates an input protoype filter into a bank of
    cosine modulated filters
    Parameters
    ----------
    h: torch.Tensor
        prototype filter
    n_band: int
        number of sub-bands
    """
    k = torch.arange(n_band).reshape(-1, 1)
    N = h.shape[-1]
    t = torch.arange(-(N // 2), N // 2 + 1)

    p = (-1)**k * math.pi / 4

    mod = torch.cos((2 * k + 1) * math.pi / (2 * n_band) * t + p)
    hk = 2 * h * mod

    return hk


def kaiser_filter(wc, atten, N=None):
    """
    Computes a kaiser lowpass filter
    Parameters
    ----------
    wc: float
        Angular frequency
    
    atten: float
        Attenuation (dB, positive)
    """
    N_, beta = kaiserord(atten, wc / np.pi)
    N_ = 2 * (N_ // 2) + 1
    N = N if N is not None else N_
    h = firwin(N, wc, window=('kaiser', beta), scale=False, nyq=np.pi)
    return h


def loss_wc(wc, atten, M, N):
    """
    Computes the objective described in https://ieeexplore.ieee.org/document/681427
    """
    h = kaiser_filter(wc, atten, N)
    g = np.convolve(h, h[::-1], "full")
    g = abs(g[g.shape[-1] // 2::2 * M][1:])
    return np.max(g)


def get_prototype(atten, M, N=None):
    """
    Given an attenuation objective and the number of bands
    returns the corresponding lowpass filter
    """
    wc = fmin(lambda w: loss_wc(w, atten, M, N), 1 / M, disp=0)[0]
    return kaiser_filter(wc, atten, N)


def polyphase_forward(x, hk, rearrange_filter=True):
    """
    Polyphase implementation of the analysis process (fast)
    Parameters
    ----------
    x: torch.Tensor
        signal to analyse ( B x 1 x T )
    
    hk: torch.Tensor
        filter bank ( M x T )
    """
    x = rearrange(x, "b c (t m) -> b (c m) t", m=hk.shape[0])
    if rearrange_filter:
        hk = rearrange(hk, "c (t m) -> c m t", m=hk.shape[0])
    x = nn.functional.conv1d(x, hk, padding=hk.shape[-1] // 2)[..., :-1]
    return x


def polyphase_inverse(x, hk, rearrange_filter=True):
    """
    Polyphase implementation of the synthesis process (fast)
    Parameters
    ----------
    x: torch.Tensor
        signal to synthesize from ( B x 1 x T )
    
    hk: torch.Tensor
        filter bank ( M x T )
    """

    m = hk.shape[0]

    if rearrange_filter:
        hk = hk.flip(-1)
        hk = rearrange(hk, "c (t m) -> m c t", m=m)  # polyphase

    pad = hk.shape[-1] // 2 + 1
    x = nn.functional.conv1d(x, hk, padding=int(pad))[..., :-1] * m

    x = x.flip(1)
    x = rearrange(x, "b (c m) t -> b c (t m)", m=m)
    x = x[..., 2 * hk.shape[1]:]
    return x


def classic_forward(x, hk):
    """
    Naive implementation of the analysis process (slow)
    Parameters
    ----------
    x: torch.Tensor
        signal to analyse ( B x 1 x T )
    
    hk: torch.Tensor
        filter bank ( M x T )
    """
    x = nn.functional.conv1d(
        x,
        hk.unsqueeze(1),
        stride=hk.shape[0],
        padding=hk.shape[-1] // 2,
    )[..., :-1]
    return x


def classic_inverse(x, hk):
    """
    Naive implementation of the synthesis process (slow)
    Parameters
    ----------
    x: torch.Tensor
        signal to synthesize from ( B x 1 x T )
    
    hk: torch.Tensor
        filter bank ( M x T )
    """
    hk = hk.flip(-1)
    y = torch.zeros(*x.shape[:2], hk.shape[0] * x.shape[-1]).to(x)
    y[..., ::hk.shape[0]] = x * hk.shape[0]
    y = nn.functional.conv1d(
        y,
        hk.unsqueeze(0),
        padding=hk.shape[-1] // 2,
    )[..., 1:]
    return y


class PQMF(nn.Module):
    """
    Pseudo Quadrature Mirror Filter multiband decomposition / reconstruction
    Parameters
    ----------
    attenuation: int
        Attenuation of the rejected bands (dB, 80 - 120)
    n_band: int
        Number of bands, must be a power of 2 if the polyphase implementation
        is needed
    """

    def __init__(self, attenuation, n_band, polyphase=True):
        super().__init__()
        h = get_prototype(attenuation, n_band)

        if polyphase:
            power = math.log2(n_band)
            assert power == math.floor(
                power
            ), "when using the polyphase algorithm, n_band must be a power of 2"

        h = torch.from_numpy(h).float()
        hk = get_qmf_bank(h, n_band)
        hk = center_pad_next_pow_2(hk)

        self.register_buffer("hk", hk)
        self.register_buffer("h", h)
        self.n_band = n_band
        self.polyphase = polyphase

    def forward(self, x):
        if self.n_band == 1:
            return x
        elif self.polyphase:
            x = polyphase_forward(x, self.hk)
        else:
            x = classic_forward(x, self.hk)

        x = reverse_half(x)

        return x

    def inverse(self, x):
        if self.n_band == 1:
            return x

        x = reverse_half(x)

        if self.polyphase:
            return polyphase_inverse(x, self.hk)
        else:
            return classic_inverse(x, self.hk)


class CachedPQMF(PQMF):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        hkf = make_odd(self.hk).unsqueeze(1)

        hki = self.hk.flip(-1)
        hki = rearrange(hki, "c (t m) -> m c t", m=self.hk.shape[0])
        hki = make_odd(hki)

        self.forward_conv = cc.Conv1d(
            hkf.shape[1],
            hkf.shape[0],
            hkf.shape[2],
            padding=cc.get_padding(hkf.shape[-1]),
            stride=hkf.shape[0],
            bias=False,
        )
        self.forward_conv.weight.data.copy_(hkf)

        self.inverse_conv = cc.Conv1d(
            hki.shape[1],
            hki.shape[0],
            hki.shape[-1],
            padding=cc.get_padding(hki.shape[-1]),
            bias=False,
        )
        self.inverse_conv.weight.data.copy_(hki)

    def script_cache(self):
        self.forward_conv.script_cache()
        self.inverse_conv.script_cache()

    def forward(self, x):
        x = self.forward_conv(x)
        x = reverse_half(x)
        return x

    def inverse(self, x):
        x = reverse_half(x)
        m = self.hk.shape[0]
        x = self.inverse_conv(x) * m
        x = x.flip(1)
        x = x.permute(0, 2, 1)
        x = x.reshape(x.shape[0], x.shape[1], -1, m).permute(0, 2, 1, 3)
        x = x.reshape(x.shape[0], x.shape[1], -1)
        return x



# Code adapted from https://github.com/lucidrains/vector-quantize-pytorch

from typing import Any, Callable, Optional, Union

import torch
import torch.nn.functional as F
from einops import repeat
from torch import nn


def ema_inplace(moving_avg, new, decay: float):
    moving_avg.data.mul_(decay).add_(new, alpha=(1 - decay))


def laplace_smoothing(x, n_categories: int, epsilon: float = 1e-5):
    return (x + epsilon) / (x.sum() + n_categories * epsilon)


def uniform_init(*shape: int):
    t = torch.empty(shape)
    nn.init.kaiming_uniform_(t)
    return t


def sample_vectors(samples, num: int):
    num_samples, device = samples.shape[0], samples.device

    if num_samples >= num:
        indices = torch.randperm(num_samples, device=device)[:num]
    else:
        indices = torch.randint(0, num_samples, (num, ), device=device)

    return samples[indices]


def kmeans(samples, num_clusters: int, num_iters: int = 10):
    dim, dtype = samples.shape[-1], samples.dtype

    means = sample_vectors(samples, num_clusters)

    for _ in range(num_iters):
        diffs = samples[:, None] - means[None]
        dists = -(diffs**2).sum(dim=-1)

        buckets = dists.max(dim=-1).indices
        bins = torch.bincount(buckets, minlength=num_clusters)
        zero_mask = bins == 0
        bins_min_clamped = bins.masked_fill(zero_mask, 1)

        new_means = buckets.new_zeros(num_clusters, dim, dtype=dtype)
        new_means.scatter_add_(0, repeat(buckets, "n -> n d", d=dim), samples)
        new_means = new_means / bins_min_clamped[..., None]

        means = torch.where(zero_mask[..., None], means, new_means)

    return means, bins


class EuclideanCodebook(nn.Module):
    """Codebook with Euclidean distance.
    Args:
        dim (int): Dimension.
        codebook_size (int): Codebook size.
        kmeans_init (bool): Whether to use k-means to initialize the codebooks.
            If set to true, run the k-means algorithm on the first training batch and use
            the learned centroids as initialization.
        kmeans_iters (int): Number of iterations used for k-means algorithm at initialization.
        decay (float): Decay for exponential moving average over the codebooks.
        epsilon (float): Epsilon value for numerical stability.
        threshold_ema_dead_code (int): Threshold for dead code expiration. Replace any codes
            that have an exponential moving average cluster size less than the specified threshold with
            randomly selected vector from the current batch.
    """

    def __init__(
        self,
        dim: int,
        codebook_size: int,
        kmeans_init: int = False,
        kmeans_iters: int = 10,
        decay: float = 0.99,
        epsilon: float = 1e-5,
        threshold_ema_dead_code: int = 2,
    ):
        super().__init__()
        self.decay = decay
        init_fn: Union[Callable[..., torch.Tensor],
                       Any] = uniform_init if not kmeans_init else torch.zeros
        embed = init_fn(codebook_size, dim)

        self.codebook_size = codebook_size

        self.kmeans_iters = kmeans_iters
        self.epsilon = epsilon
        self.threshold_ema_dead_code = threshold_ema_dead_code

        self.register_buffer("inited", torch.Tensor([not kmeans_init]))
        self.register_buffer("cluster_size", torch.zeros(codebook_size))
        self.register_buffer("embed", embed)
        self.register_buffer("embed_avg", embed.clone())

    @torch.jit.unused
    def init_embed_(self, data):
        embed, cluster_size = kmeans(data, self.codebook_size,
                                     self.kmeans_iters)
        self.embed.data.copy_(embed)
        self.embed_avg.data.copy_(embed.clone())
        self.cluster_size.data.copy_(cluster_size)
        self.inited.data.copy_(torch.Tensor([True]))

    def replace_(self, samples, mask):
        modified_codebook = torch.where(
            mask[..., None], sample_vectors(samples, self.codebook_size),
            self.embed)
        self.embed.data.copy_(modified_codebook)

    def expire_codes_(self, batch_samples):
        if self.threshold_ema_dead_code == 0:
            return

        expired_codes = self.cluster_size < self.threshold_ema_dead_code
        if not torch.any(expired_codes):
            return

        batch_samples = batch_samples.reshape(-1, batch_samples.shape[-1])
        self.replace_(batch_samples, mask=expired_codes)

    def preprocess(self, x):
        return x.reshape(-1, x.shape[-1])

    def quantize(self, x):
        embed = self.embed.t()
        dist = -(x.pow(2).sum(1, keepdim=True) - 2 * x @ embed +
                 embed.pow(2).sum(0, keepdim=True))
        embed_ind = dist.max(dim=-1).indices
        return embed_ind

    def dequantize(self, embed_ind):
        quantize = F.embedding(embed_ind, self.embed)
        return quantize

    def encode(self, x):
        shape = x.shape
        # pre-process
        x = self.preprocess(x)
        # quantize
        embed_ind = self.quantize(x)
        # post-process
        embed_ind = embed_ind.reshape(shape[0], shape[1])
        return embed_ind

    def decode(self, embed_ind):
        quantize = self.dequantize(embed_ind)
        return quantize

    def forward(self, x):
        shape, dtype = x.shape, x.dtype
        x = self.preprocess(x)

        if not self.inited:
            self.init_embed_(x)

        embed_ind = self.quantize(x)
        embed_onehot = F.one_hot(embed_ind, self.codebook_size).type(dtype)
        embed_ind = embed_ind.reshape(shape[0], shape[1])
        quantize = self.dequantize(embed_ind)

        if self.training:
            # We do the expiry of code at that point as buffers are in sync
            # and all the workers will take the same decision.
            self.expire_codes_(x)
            ema_inplace(self.cluster_size, embed_onehot.sum(0), self.decay)
            embed_sum = x.t() @ embed_onehot
            ema_inplace(self.embed_avg, embed_sum.t(), self.decay)
            cluster_size = (laplace_smoothing(
                self.cluster_size, self.codebook_size, self.epsilon) *
                            self.cluster_size.sum())
            embed_normalized = self.embed_avg / cluster_size.unsqueeze(1)
            self.embed.data.copy_(embed_normalized)

        return quantize, embed_ind


class VectorQuantization(nn.Module):
    """Vector quantization implementation.
    Currently supports only euclidean distance.
    Args:
        dim (int): Dimension
        codebook_size (int): Codebook size
        codebook_dim (int): Codebook dimension. If not defined, uses the specified dimension in dim.
        decay (float): Decay for exponential moving average over the codebooks.
        epsilon (float): Epsilon value for numerical stability.
        kmeans_init (bool): Whether to use kmeans to initialize the codebooks.
        kmeans_iters (int): Number of iterations used for kmeans initialization.
        threshold_ema_dead_code (int): Threshold for dead code expiration. Replace any codes
            that have an exponential moving average cluster size less than the specified threshold with
            randomly selected vector from the current batch.
        commitment_weight (float): Weight for commitment loss.
    """

    def __init__(
        self,
        dim: int,
        codebook_size: int,
        codebook_dim: Optional[int] = None,
        decay: float = 0.99,
        epsilon: float = 1e-5,
        kmeans_init: bool = True,
        kmeans_iters: int = 50,
        threshold_ema_dead_code: int = 2,
        commitment_weight: float = 1.,
    ):
        super().__init__()
        _codebook_dim: int = codebook_dim or dim

        requires_projection = _codebook_dim != dim
        self.project_in = (nn.Linear(dim, _codebook_dim)
                           if requires_projection else nn.Identity())
        self.project_out = (nn.Linear(_codebook_dim, dim)
                            if requires_projection else nn.Identity())

        self.epsilon = epsilon
        self.commitment_weight = commitment_weight

        self._codebook = EuclideanCodebook(
            dim=_codebook_dim,
            codebook_size=codebook_size,
            kmeans_init=kmeans_init,
            kmeans_iters=kmeans_iters,
            decay=decay,
            epsilon=epsilon,
            threshold_ema_dead_code=threshold_ema_dead_code)
        self.codebook_size = codebook_size

    @property
    def codebook(self):
        return self._codebook.embed

    def encode(self, x):
        x = x.permute(0, 2, 1)
        x = self.project_in(x)
        embed_in = self._codebook.encode(x)
        return embed_in

    def decode(self, embed_ind):
        quantize = self._codebook.decode(embed_ind)
        quantize = self.project_out(quantize)
        quantize = quantize.permute(0, 2, 1)
        return quantize

    def forward(self, x):
        device = x.device
        x = x.permute(0, 2, 1)
        x = self.project_in(x)

        quantize, embed_ind = self._codebook(x)

        if self.training:
            quantize = x + (quantize - x).detach()

        loss = torch.tensor([0.0], device=device, requires_grad=self.training)

        if self.training:
            if self.commitment_weight > 0:
                commit_loss = F.mse_loss(quantize.detach(), x)
                loss = loss + commit_loss * self.commitment_weight

        quantize = self.project_out(quantize)
        quantize = quantize.permute(0, 2, 1)
        return quantize, embed_ind, loss


class ResidualVectorQuantization(nn.Module):
    """Residual vector quantization implementation.
    Follows Algorithm 1. in https://arxiv.org/pdf/2107.03312.pdf
    """

    def __init__(self, num_quantizers, **kwargs):
        super().__init__()
        self.layers = nn.ModuleList(
            [VectorQuantization(**kwargs) for _ in range(num_quantizers)])

    def forward(self, x):
        quantized_out = 0.0
        residual = x

        all_losses = []
        all_indices = []

        for layer in self.layers:
            quantized, indices, loss = layer(residual)
            residual = residual - quantized
            quantized_out = quantized_out + quantized

            all_indices.append(indices)
            all_losses.append(loss)

        out_losses = torch.stack(all_losses, 0).sum()
        all_indices = torch.stack(all_indices, 1)
        return quantized_out, out_losses, all_indices

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        all_indices = []
        for layer in self.layers:
            indices = layer.encode(residual)
            quantized = layer.decode(indices)
            residual = residual - quantized
            all_indices.append(indices)
        out_indices = torch.stack(all_indices, 1)
        return out_indices

    def decode(self, q_indices: torch.Tensor) -> torch.Tensor:
        quantized_out = torch.tensor(0.0, device=q_indices.device)
        for i, layer in enumerate(self.layers):
            quantized = layer.decode(q_indices[:, i])
            quantized_out = quantized_out + quantized
        return quantized_out


# Balancer - adapted from https://github.com/facebookresearch/encodec

from typing import Any, Callable, Dict, Optional, Sequence

import torch


class EMA:

    def __init__(self, beta: float = .999) -> None:
        self.shadows = {}
        self.beta = beta

    def __call__(self, inputs: Dict[str, torch.Tensor]):
        outputs = {}
        for k, v in inputs.items():
            if not k in self.shadows:
                self.shadows[k] = v
            else:
                self.shadows[k] *= self.beta
                self.shadows[k] += (1 - self.beta) * v

            outputs[k] = self.shadows[k].clone()
        return outputs


class Balancer:

    def __init__(
        self,
        ema_averager: Callable[[], EMA],
        weights: Dict[str, float],
        scale_gradients: bool = False,
        deny_list: Optional[Sequence[str]] = None,
    ) -> None:
        self.ema_averager = ema_averager()
        self.weights = weights
        self.scale_gradients = scale_gradients
        self.deny_list = deny_list

    def backward(
        self,
        losses: Dict[str, torch.Tensor],
        model_output: torch.Tensor,
        logger: Optional[Callable[[str, float], None]] = None,
        profiler: Optional[Any] = None,
    ):
        grads = {}
        norms = {}

        for k, v in losses.items():
            if self.deny_list is not None:
                if k in self.deny_list: continue

            grads[k], = torch.autograd.grad(
                v,
                [model_output],
                retain_graph=True,
            )

            if (nans := torch.isnan(grads[k])).any():
                count = nans.float().mean()
                grads[k] = torch.where(nans, torch.zeros_like(nans), grads[k])
                if logger is not None:
                    logger(f"{k}_nan_ratio", count)

            norms[k] = grads[k].norm(
                dim=tuple(range(1, grads[k].dim()))).mean()

            if profiler is not None:
                profiler(f'partial backward {k}')

        avg_norms = self.ema_averager(norms)

        if profiler is not None:
            profiler('grad norm estimation')

        sum_weights = sum([self.weights.get(k, 1) for k in avg_norms])

        for name, norm in avg_norms.items():
            if self.scale_gradients:
                ratio = self.weights.get(name, 1) / sum_weights
                scale = ratio / (norm + 1e-6)
                grads[name] *= scale

                if logger is not None:
                    logger(f'scale_{name}', scale)
                    logger(f'grad_norm_{name}', grads[name].norm())
                    logger(f'target_norm_{name}', ratio)
            else:
                scale = self.weights.get(name, 1)
                grads[name] *= scale

                if logger is not None:
                    logger(f'scale_{name}', scale)
                    logger(f'grad_norm_{name}', grads[name].norm())

        if profiler is not None:
            profiler('norm scaling')

        full_grad = sum([grads[name] for name in avg_norms.keys()])
        model_output.backward(full_grad, retain_graph=True)

        if profiler is not None:
            profiler('scaled backward')

        if self.deny_list is not None:
            for k in self.deny_list:
                if k in losses:
                    loss = losses[k] * self.weights.get(k, 1)
                    if logger is not None:
                        logger(f'scale_{name}', scale)
                        logger(f'grad_norm_{name}', grads[name].norm())
                    if loss.requires_grad:
                        loss.backward(retain_graph=True)

        if profiler is not None:
            profiler('denied backward')



import torch

torch.set_grad_enabled(False)
import os

import cached_conv as cc
import gin
import torch.nn as nn
from absl import app, flags
from effortless_config import Config

import rave

flags.DEFINE_string('run', default=None, required=True, help='Run to export')
FLAGS = flags.FLAGS


def main(argv):
    gin.parse_config_file(os.path.join(FLAGS.run, "config.gin"))
    checkpoint = rave.core.search_for_run(FLAGS.run)

    print(f"using {checkpoint}")

    pretrained = rave.RAVE()
    pretrained.load_state_dict(torch.load(checkpoint)["state_dict"])
    pretrained.eval()

    for m in pretrained.modules():
        if hasattr(m, "weight_g"):
            nn.utils.remove_weight_norm(m)

    def recursive_replace(model: nn.Module):
        for name, child in model.named_children():
            if isinstance(child, cc.convs.Conv1d):
                conv = nn.Conv1d(
                    child.in_channels,
                    child.out_channels,
                    child.kernel_size,
                    child.stride,
                    child._pad[0],
                    child.dilation,
                    child.groups,
                    child.bias,
                )
                conv.weight.data.copy_(child.weight.data)
                if conv.bias is not None:
                    conv.bias.data.copy_(child.bias.data)
                setattr(model, name, conv)
            elif isinstance(child, cc.convs.ConvTranspose1d):
                conv = nn.ConvTranspose1d(
                    child.in_channels,
                    child.out_channels,
                    child.kernel_size,
                    child.stride,
                    child.padding,
                    child.output_padding,
                    child.groups,
                    child.bias,
                    child.dilation,
                    child.padding_mode,
                )
                conv.weight.data.copy_(child.weight.data)
                if conv.bias is not None:
                    conv.bias.data.copy_(child.bias.data)
                setattr(model, name, conv)
            else:
                recursive_replace(child)

    recursive_replace(pretrained)

    x = torch.randn(1, 1, 2**15)
    pretrained(x)

    name = os.path.basename(os.path.normpath(FLAGS.run))
    export_path = os.path.join(FLAGS.run, name)
    torch.onnx.export(
        pretrained,
        x,
        f"{export_path}.onnx",
        export_params=True,
        opset_version=12,
        input_names=["audio_in"],
        output_names=["audio_out"],
        dynamic_axes={
            "audio_in": {
                2: "audio_length"
            },
            "audio_out": [0],
        },
        do_constant_folding=False,
    )


if __name__ == '__main__':
    app.run(main)


from time import time
from typing import Callable, Optional

import gin
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
from einops import rearrange
from sklearn.decomposition import PCA

import rave.core

from .balancer import Balancer
from .blocks import DiscreteEncoder, VariationalEncoder


class Profiler:

    def __init__(self):
        self.ticks = [[time(), None]]

    def tick(self, msg):
        self.ticks.append([time(), msg])

    def __repr__(self):
        rep = 80 * "=" + "\n"
        for i in range(1, len(self.ticks)):
            msg = self.ticks[i][1]
            ellapsed = self.ticks[i][0] - self.ticks[i - 1][0]
            rep += msg + f": {ellapsed*1000:.2f}ms\n"
        rep += 80 * "=" + "\n\n\n"
        return rep


class WarmupCallback(pl.Callback):

    def __init__(self) -> None:
        super().__init__()
        self.state = {'training_steps': 0}

    def on_train_batch_start(self, trainer, pl_module, batch,
                             batch_idx) -> None:
        if self.state['training_steps'] >= pl_module.warmup:
            pl_module.warmed_up = True
        self.state['training_steps'] += 1

    def state_dict(self):
        return self.state.copy()

    def load_state_dict(self, state_dict):
        self.state.update(state_dict)


class QuantizeCallback(WarmupCallback):

    def on_train_batch_start(self, trainer, pl_module, batch,
                             batch_idx) -> None:

        if pl_module.warmup_quantize is None: return

        if self.state['training_steps'] >= pl_module.warmup_quantize:
            if isinstance(pl_module.encoder, DiscreteEncoder):
                pl_module.encoder.enabled = torch.tensor(1).type_as(
                    pl_module.encoder.enabled)
        self.state['training_steps'] += 1


@gin.configurable
class RAVE(pl.LightningModule):

    def __init__(
        self,
        latent_size,
        sampling_rate,
        encoder,
        decoder,
        discriminator,
        phase_1_duration,
        gan_loss,
        valid_signal_crop,
        feature_matching_fun,
        num_skipped_features,
        audio_distance: Callable[[], nn.Module],
        multiband_audio_distance: Callable[[], nn.Module],
        balancer: Callable[[], Balancer],
        warmup_quantize: Optional[int] = None,
        pqmf: Optional[Callable[[], nn.Module]] = None,
        update_discriminator_every: int = 2,
    ):
        super().__init__()

        self.pqmf = None
        if pqmf is not None:
            self.pqmf = pqmf()

        self.encoder = encoder()
        self.decoder = decoder()
        self.discriminator = discriminator()

        self.audio_distance = audio_distance()
        self.multiband_audio_distance = multiband_audio_distance()

        self.gan_loss = gan_loss

        self.register_buffer("latent_pca", torch.eye(latent_size))
        self.register_buffer("latent_mean", torch.zeros(latent_size))
        self.register_buffer("fidelity", torch.zeros(latent_size))

        self.latent_size = latent_size

        self.automatic_optimization = False

        # SCHEDULE
        self.warmup = phase_1_duration
        self.warmup_quantize = warmup_quantize
        self.balancer = balancer()

        self.warmed_up = False

        # CONSTANTS
        self.sr = sampling_rate
        self.valid_signal_crop = valid_signal_crop
        self.feature_matching_fun = feature_matching_fun
        self.num_skipped_features = num_skipped_features
        self.update_discriminator_every = update_discriminator_every

        self.eval_number = 0
        self.integrator = None

        self.register_buffer("receptive_field", torch.tensor([0, 0]).long())

    def configure_optimizers(self):
        gen_p = list(self.encoder.parameters())
        gen_p += list(self.decoder.parameters())
        dis_p = list(self.discriminator.parameters())

        gen_opt = torch.optim.Adam(gen_p, 1e-4, (.5, .9))
        dis_opt = torch.optim.Adam(dis_p, 1e-4, (.5, .9))

        return gen_opt, dis_opt

    def split_features(self, features):
        feature_real = []
        feature_fake = []
        for scale in features:
            true, fake = zip(*map(
                lambda x: torch.split(x, x.shape[0] // 2, 0),
                scale,
            ))
            feature_real.append(true)
            feature_fake.append(fake)
        return feature_real, feature_fake

    def training_step(self, batch, batch_idx):
        p = Profiler()
        gen_opt, dis_opt = self.optimizers()
        x = batch.unsqueeze(1)

        if self.pqmf is not None:
            x_multiband = self.pqmf(x)
        else:
            x_multiband = x
        p.tick('decompose')

        self.encoder.set_warmed_up(self.warmed_up)
        self.decoder.set_warmed_up(self.warmed_up)

        # ENCODE INPUT
        z_pre_reg = self.encoder(x_multiband)
        z, reg = self.encoder.reparametrize(z_pre_reg)[:2]
        p.tick('encode')

        # DECODE LATENT
        y_multiband = self.decoder(z)
        p.tick('decode')

        if self.valid_signal_crop and self.receptive_field.sum():
            x_multiband = rave.core.valid_signal_crop(
                x_multiband,
                *self.receptive_field,
            )
            y_multiband = rave.core.valid_signal_crop(
                y_multiband,
                *self.receptive_field,
            )
        p.tick('crop')

        # DISTANCE BETWEEN INPUT AND OUTPUT
        distances = {}

        if self.pqmf is not None:
            multiband_distance = self.multiband_audio_distance(
                x_multiband, y_multiband)
            p.tick('mb distance')

            x = self.pqmf.inverse(x_multiband)
            y = self.pqmf.inverse(y_multiband)
            p.tick('recompose')

            for k, v in multiband_distance.items():
                distances[f'multiband_{k}'] = v
        else:
            x = x_multiband
            y = y_multiband

        fullband_distance = self.audio_distance(x, y)
        p.tick('fb distance')

        for k, v in fullband_distance.items():
            distances[f'fullband_{k}'] = v

        feature_matching_distance = 0.

        if self.warmed_up:  # DISCRIMINATION
            xy = torch.cat([x, y], 0)
            features = self.discriminator(xy)

            feature_real, feature_fake = self.split_features(features)

            loss_dis = 0
            loss_adv = 0

            pred_real = 0
            pred_fake = 0

            for scale_real, scale_fake in zip(feature_real, feature_fake):
                current_feature_distance = sum(
                    map(
                        self.feature_matching_fun,
                        scale_real[self.num_skipped_features:],
                        scale_fake[self.num_skipped_features:],
                    )) / len(scale_real[self.num_skipped_features:])

                feature_matching_distance = feature_matching_distance + current_feature_distance

                _dis, _adv = self.gan_loss(scale_real[-1], scale_fake[-1])

                pred_real = pred_real + scale_real[-1].mean()
                pred_fake = pred_fake + scale_fake[-1].mean()

                loss_dis = loss_dis + _dis
                loss_adv = loss_adv + _adv

            feature_matching_distance = feature_matching_distance / len(
                feature_real)

        else:
            pred_real = torch.tensor(0.).to(x)
            pred_fake = torch.tensor(0.).to(x)
            loss_dis = torch.tensor(0.).to(x)
            loss_adv = torch.tensor(0.).to(x)
        p.tick('discrimination')

        # COMPOSE GEN LOSS
        loss_gen = {}
        loss_gen.update(distances)
        p.tick('update loss gen dict')

        if reg.item():
            loss_gen['regularization'] = reg

        if self.warmed_up:
            loss_gen['feature_matching'] = feature_matching_distance
            loss_gen['adversarial'] = loss_adv

        # OPTIMIZATION
        if not (batch_idx %
                self.update_discriminator_every) and self.warmed_up:
            dis_opt.zero_grad()
            loss_dis.backward()
            dis_opt.step()
            p.tick('dis opt')
        else:
            gen_opt.zero_grad()
            self.balancer.backward(loss_gen, y_multiband, self.log, p.tick)
            gen_opt.step()

        # LOGGING
        if self.warmed_up:
            self.log("loss_dis", loss_dis)
            self.log("pred_real", pred_real.mean())
            self.log("pred_fake", pred_fake.mean())

        self.log_dict(loss_gen)
        p.tick('logging')

    def encode(self, x):
        if self.pqmf is not None:
            x = self.pqmf(x)
        z, = self.encoder.reparametrize(self.encoder(x))[:1]
        return z

    def decode(self, z):
        y = self.decoder(z)
        if self.pqmf is not None:
            y = self.pqmf.inverse(y)
        return y

    def forward(self, x):
        return self.decode(self.encode(x))

    def validation_step(self, batch, batch_idx):

        x = batch.unsqueeze(1)

        if self.pqmf is not None:
            x = self.pqmf(x)

        z = self.encoder(x)

        if isinstance(self.encoder, VariationalEncoder):
            mean = torch.split(z, z.shape[1] // 2, 1)[0]
        else:
            mean = None

        z = self.encoder.reparametrize(z)[0]
        y = self.decoder(z)

        if self.pqmf is not None:
            x = self.pqmf.inverse(x)
            y = self.pqmf.inverse(y)

        distance = self.audio_distance(x, y)

        full_distance = sum(distance.values())

        if self.trainer is not None:
            self.log('validation', full_distance)

        return torch.cat([x, y], -1), mean

    def validation_epoch_end(self, out):
        if not self.receptive_field.sum():
            print("Computing receptive field for this configuration...")
            lrf, rrf = rave.core.get_rave_receptive_field(self)
            self.receptive_field[0] = lrf
            self.receptive_field[1] = rrf
            print(
                f"Receptive field: {1000*lrf/self.sr:.2f}ms <-- x --> {1000*rrf/self.sr:.2f}ms"
            )

        if not len(out): return

        audio, z = list(zip(*out))
        audio = list(map(lambda x: x.cpu(), audio))

        # LATENT SPACE ANALYSIS
        if not self.warmed_up and isinstance(self.encoder, VariationalEncoder):
            z = torch.cat(z, 0)
            z = rearrange(z, "b c t -> (b t) c")

            self.latent_mean.copy_(z.mean(0))
            z = z - self.latent_mean

            pca = PCA(z.shape[-1]).fit(z.cpu().numpy())

            components = pca.components_
            components = torch.from_numpy(components).to(z)
            self.latent_pca.copy_(components)

            var = pca.explained_variance_ / np.sum(pca.explained_variance_)
            var = np.cumsum(var)

            self.fidelity.copy_(torch.from_numpy(var).to(self.fidelity))

            var_percent = [.8, .9, .95, .99]
            for p in var_percent:
                self.log(
                    f"fidelity_{p}",
                    np.argmax(var > p).astype(np.float32),
                )

        y = torch.cat(audio, 0)[:8].reshape(-1).numpy()

        if self.integrator is not None:
            y = self.integrator(y)

        self.logger.experiment.add_audio("audio_val", y, self.eval_number,
                                         self.sr)
        self.eval_number += 1

    def on_fit_start(self):
        tb = self.logger.experiment

        config = gin.operative_config_str()
        config = config.split('\n')
        config = ['```'] + config + ['```']
        config = '\n'.join(config)
        tb.add_text("config", config)

        model = str(self)
        model = model.split('\n')
        model = ['```'] + model + ['```']
        model = '\n'.join(model)
        tb.add_text("model", model)



import itertools

import cached_conv as cc
import pytest
import torch

from rave.blocks import *

kernel_size = [
    1,
    3,
]

dilations = [[1, 1], [3, 1]]

kernel_sizes = [
    [3],
    [3, 5],
    [3, 5, 7],
]

dilations_list = [
    [[1, 1]],
    [[1, 1], [3, 1], [5, 1]],
]

ratios = [
    2,
    4,
    8,
]


@pytest.mark.parametrize('kernel_sizes,dilations_list',
                         itertools.product(kernel_sizes, dilations_list))
def test_residual_stack(kernel_sizes, dilations_list):
    dim = 16
    x = torch.randn(1, dim, 32)
    cc.use_cached_conv(False)
    stack_regular = ResidualStack(
        dim=dim,
        kernel_sizes=[3],
        dilations_list=[[1, 1], [3, 1], [5, 1]],
    )

    cc.use_cached_conv(True)
    stack_stream = ResidualStack(
        dim=dim,
        kernel_sizes=[3],
        dilations_list=[[1, 1], [3, 1], [5, 1]],
    )

    for p1, p2 in zip(stack_regular.parameters(), stack_stream.parameters()):
        p2.data.copy_(p1.data)

    delay = stack_stream.cumulative_delay

    y_regular = stack_regular(x)
    y_stream = stack_stream(x)

    if delay:
        y_regular = y_regular[..., delay:-delay]
        y_stream = y_stream[..., delay + delay:]

    assert torch.allclose(y_regular, y_stream, 1e-4, 1e-4)


@pytest.mark.parametrize('kernel_size,dilations_list',
                         itertools.product(kernel_size, dilations))
def test_residual_layer(kernel_size, dilations_list):
    dim = 16
    x = torch.randn(1, dim, 32)

    cc.use_cached_conv(False)
    layer_regular = ResidualLayer(dim, kernel_size, dilations_list)

    cc.use_cached_conv(True)
    layer_stream = ResidualLayer(dim, kernel_size, dilations_list)

    for p1, p2 in zip(layer_regular.parameters(), layer_stream.parameters()):
        p2.data.copy_(p1.data)

    delay = layer_stream.cumulative_delay

    y_regular = layer_regular(x)
    y_stream = layer_stream(x)

    if delay:
        y_regular = y_regular[..., delay:-delay]
        y_stream = y_stream[..., delay + delay:]

    assert torch.allclose(y_regular, y_stream, 1e-3, 1e-4)


@pytest.mark.parametrize('ratio,', ratios)
def test_upsample_layer(ratio):
    dim = 16
    x = torch.randn(1, dim, 32)

    cc.use_cached_conv(False)
    upsample_regular = UpsampleLayer(dim, dim, ratio)

    cc.use_cached_conv(True)
    upsample_stream = UpsampleLayer(dim, dim, ratio)

    for p1, p2 in zip(upsample_regular.parameters(),
                      upsample_stream.parameters()):
        p2.data.copy_(p1.data)

    delay = upsample_stream.cumulative_delay

    y_regular = upsample_regular(x)
    y_stream = upsample_stream(x)

    if delay:
        y_regular = y_regular[..., delay:-delay]
        y_stream = y_stream[..., delay + delay:]

    assert torch.allclose(y_regular, y_stream, 1e-3, 1e-4)
